"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import sys
from pathlib import Path

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


def cli_show_info(argv=None):
    argv = argv or sys.argv[1:]
# region [Main_Exec]


if __name__ == '__main__':
    cli_show_info()

# endregion [Main_Exec]
