"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from pathlib import Path
from rich.console import Console as RichConsole
from rich.style import Style
from rich.default_styles import DEFAULT_STYLES
# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


# region [Main_Exec]

if __name__ == '__main__':
    pass

# endregion [Main_Exec]
