.aboutSummary {
    background-color: lightgrey;
    text-align: center;
    color: black;
    font-style: italic;
    font-weight: bold;
    font-size: large;
}

dd {
    float: center;
    padding: 50rem;
}